package NNIPF;
#
#This file is part of NNIPF(Version 0.06-release):SPAM mail filtering software
#Copyright (C) Toshikazu Wada <twada@ieee.org>
#
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
use strict;
use warnings;
my $BADIP;
my $GOODIP;
my $MBADIP;
my $MGOODIP;

#################
sub SetDict{
    ($BADIP,$GOODIP,$MBADIP,$MGOODIP)=@_;
}
###########################
sub IpDictAddEG{
  my ($OIP) = @_;
  IpDictAddExclusive($OIP,$GOODIP,$BADIP);
}
sub IpDictAddEB{
  my ($OIP) = @_;
  IpDictAddExclusive($OIP,$BADIP,$GOODIP);
}
sub IpDictAddB{
  my ($OIP) = @_;
  IpDictAdd($OIP,$BADIP);
}
sub IpDictAddG{
  my ($OIP) = @_;
  IpDictAdd($OIP,$GOODIP);
}
#################
sub IsPrivate{
  my ($OIP) = @_;
  my $res = 0;
  if (($OIP =~ /^10\.[0-9]+\.[0-9]+\.[0-9]+/)||
      ($OIP =~ /^172\.(16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)\.[0-9]+\.[0-9]+/)||
      ($OIP =~ /^192\.168\.[0-9]+\.[0-9]+/ )) {$res=1;}
  return $res;
}

sub min{
  my ($a, $b) = @_;
  return ($a<$b)? $a:$b;
}

################
sub IpDictMatch{
  my ($query,$Dict, $Flg) = @_;
  my @qaddr=split(/\./,$query);
  my $Dir= defined($Dict) ? $Dict: "";
  my $fc=0;
  my $count;
  my $counter;
  my $flag;
  my $filename;
  my $file;
  my $i;
  my $match;
  
  if (-d $Dir) {
    for ($i=0, $match=0, $flag=1; $i<4; $i++){
      if (-d ($filename="$Dir/$qaddr[$i]") ){
        $fc=0;
        if ($Flg){
          if (-e ($counter="$filename/Count")){
            open INc, $counter or die "$counter:$!";
            while(<INc>) {$count=$_;}
            close(INc);
            unlink($counter);
            $count+=1;
          }else{$count=1;}
          open OUTc, ">$counter" or die "$counter:$!";
          print(OUTc $count);
          close(OUTc);
        }
      }elsif ($flag) {
        $flag=0;
        opendir DH, $Dir or die "$Dir:$!";
        $fc=256;
        while ($file = readdir DH) {  
          next if $file =~ /^\.{1,2}$/;
          if (abs(str2int($file)-str2int($qaddr[$i])) < $fc) {
            $fc=abs(str2int($file)-str2int($qaddr[$i]));
          }
        }
      }else{$fc=256;}
      $match=$match*256+$fc;
      $Dir=$filename;
    }
  }else{ $match = 4294967295; }
  return $match;
}
#####################
sub str2int{
  my ($str)=@_;
  my $i;
  my $len;
  my $ret=0;
  if ($str =~ /([0-9]+)/) {
    $str=$1;
    $len = length($str);
    for ($i = 0; $i< $len; $i++) {
      $ret *= 10;
      $ret += ord(substr($str,$i,1))-ord('0');
    }
  }
  return $ret;
}
#####################
sub IpDictAdd{
  my ($query,$Dict) = @_;
  my @qaddr=split(/\./,$query);
  my $filename=$Dict;
  my $i;
  my $counter;
  my $count;
  
  for ($i=0; $i<4; $i++){
    $filename="$filename/$qaddr[$i]";
    if (!-e $filename ) {
      mkdir($filename);
    }
    
    if (-e ($counter="$filename/Count")){
      open INc, $counter or die "$counter:$!";
      while(<INc>) {$count=$_;}
      close(INc);
    }else{$count=1;}
    open OUTc, ">$counter" or die "$counter:$!";
    print(OUTc $count);
    close(OUTc);
  }
}

####################
sub IpDictRemove{
  my ($query,$Dict) = @_;
  $query =~ s/\./\//g;
  my $Dir=$Dict . "/" . $query;
  
  my $prevfc=0;
  my $prevdir=$Dir;
  my $i;
  my $fc;
  my $file;
  
  if (-d $Dir) {
    for ($i=0; $i<4; $i++){
      opendir DH, $Dir or die "$Dir:$!";
      $fc=0;
      while ($file = readdir DH) {  
        next if $file =~ /^\.{1,2}$/;
        $fc++;
        $file = $Dir . "/" . $file; 
        if ( -f $file ) {unlink($file);$fc--;}
        elsif (($file eq $prevdir)&&(-d $file)&&($prevfc eq 0))
          {rmdir($file);$fc--;}
      }
      $prevdir = $Dir;
      $prevfc = $fc;
      $Dir=substr($Dir,0,rindex($Dir,"/"));
      closedir DH;
    }
  }
}

###########################
sub IpDictAddExclusive{
  my ($OIP, $DICT1, $DICT2) = @_;
  IpDictAdd($OIP,$DICT1);
  if (IpDictMatch($OIP,$DICT2,0) eq 0) {
    IpDictRemove($OIP,$DICT2);
  }
}

######################
sub Bratio{
  my ($Dg, $Db) = @_;
  my $eps;
  
  if ($Dg+$Db eq 0) {$eps=1;}else{$eps=0;}
  return $Dg/($Dg+$Db+$eps);
}
######################
sub distinctiveness{
  my ($OIP, $Flg) = @_;
  if (IsPrivate($OIP)) {return 0;}
  else{
    ### Check IP address
    my $Bip;
    my $Gip;
    my $SBip = IpDictMatch($OIP,$BADIP,$Flg);
    my $MBip = IpDictMatch($OIP,$MBADIP,0);
    my $SGip = IpDictMatch($OIP,$GOODIP,$Flg);
    my $MGip = IpDictMatch($OIP,$MGOODIP,0);
    if ( (($SBip eq 0)&&($MGip eq 0))|| 
         (($SGip eq 0)&&($MBip eq 0)) ){
      $Bip = $SBip;
      $Gip = $SGip;
    }else{
      $Bip = min($SBip,$MBip);
      $Gip = min($SGip,$MGip);
    }
    return Bratio($Gip,$Bip);
  }
}

sub HeaderAnalysis{
  my ($INt, $Buf, $Flg) = @_;
  my $lcount=0;
  my $line="";
  my $res;
  my $rf;
  my  $distinc;
  # Init
  my $hflag=1;
  my $Rflag = 0;
  my $Maxdistinc=0;
  my $OIP;
  my $Rline;
  
  $res->{'Frm'}=""; $rf->{'rfrom'}="";
  while (($hflag)&& ($line = <$INt>)){
    $$Buf[$lcount++] = $line;
    $line =~ s/\n//;
    # When Empty line is read
    if (length($line) eq 0) {$hflag = 0; }
    # Make each Field into a single line
    if ($line =~ /^([A-z0-9\-]*): /) {
      if ($Rflag eq 1) {
        if ($Rline =~ /^Received: (.*)$/) {
          
          $rf=ReceivedFrom($1);
          if (($rf->{'by'})&&(Checkb($rf->{'by'}))) {
            if ($rf->{'rfrom'}) {
              $OIP=officialIP($rf->{'rfrom'});
              if (($OIP)&&(($distinc=distinctiveness($OIP,$Flg))>=$Maxdistinc)){
                $res->{'Frm'}=$rf->{'rfrom'};
                $res->{'OIP'}=$OIP;
                $res->{'Dis'}=$distinc;
                $res->{'by'}=$rf->{'by'};
                $Maxdistinc=$distinc;
              }
            }
          }
        }else{
          if ($Rline =~ /^From: /) {$res->{'from'} = substr($Rline,6);}
          elsif ($Rline =~ /^Subject: /) {$res->{'subject'} = substr($Rline,9);}
        }
      } else {$Rflag=1;}
      $Rline=$line;
    }else{
      if ($Rflag) {
        $line =~ s/^\s*//g;
        $Rline = $Rline . $line;
      }
    }
  }
  ####### Post Processing
  if (($res->{'Frm'} eq "")&&($rf->{'rfrom'} ne "")) {
    $res->{'Frm'}=$rf->{'rfrom'};
    $OIP=officialIP($res->{'Frm'});
    $res->{'OIP'}=$OIP;
    $res->{'Dis'}=distinctiveness($OIP,$Flg);
  }elsif ($res->{'Frm'} eq "") {
    $res->{'Frm'}='';$res->{'OIP'}='';$res->{'Dis'}=0;
  }
  $res->{'Frm'} =~ s/(.*)\n/$1/;
  $res->{'lcount'} = $lcount;
  return $res;
}

############# string => from string and by string
# hash members = 'rfrom', 'by'
#
sub ReceivedFrom{
  my ($l)=@_;
  my $res;
  if ($l =~ /^from (.*)by (.*)for (.*);(.*)$/ ){
    $res->{'rfrom'} = $1;
    $res->{'by'} = substr($2,0,index($2," with"));
    if ($res->{'by'} =~ /([A-z0-9\.\-_]*) /) {$res->{'by'} = $1;}
    $res->{'rfrom'} =~ s/ *$//;
  }elsif ($l =~ /^from (.*)by (.*);(.*)$/){
    $res->{'rfrom'} = $1;
    $res->{'by'} = substr($2,0,index($2," with"));
    if ($res->{'by'} =~ /([A-z0-9\.\-_]*) /) {
      $res->{'by'} = $1;
    }
    $res->{'rfrom'} =~ s/ *$//;
  }
  return $res;
}
######################## from string returned by ReceivedFrom() to official IP
sub officialIP{
  my ($str)=@_;
  my $OIP="";
  if ($str =~ /.*\[([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)\].*/) {
    $OIP =$1;
  }
  my @h=split(/ /,$str);
  if  ((defined $h[3])&&($h[3] =~ /[^0-9]*([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)[^0-9]*/)) {
    $OIP = $1;
  }
  return $OIP;
}
############# Reading Boundary File (side effect to %hash)
sub bRead{
  my ($Dict) = @_;
  our @hash = ();
  my $i=0;
  if (defined($Dict)&&(-e $Dict)){
    open INb, $Dict or die "$Dict:$!";
    while(<INb>){ $_ =~ s/\n//; $_ =~ s/\r//; $hash[$i++]="$_"; }
    close(INb);
  }else {exit 0;}
}

############# Check whether $query matches MTAOB or not? true or false 
sub Checkb {
  my ($query) = @_;
  my $i;
  our @hash;
  
  foreach $i (@hash) {
    if ( $query =~ /^$i$/){
      return 1;
    }
  }
  return 0;
}
sub rvf{
  my ($fn) = @_;
  my $header;
  my @Buf;
  open INtmp, $fn or die "$fn: $!";
  $header=HeaderAnalysis(\*INtmp,\@Buf,0);
  close(INtmp);
  return $header->{'Frm'};
}
1;
